/*  samples/cctrans_sample.p  Process a credit card for authorization  SAMPLE
            DY Tri-8, 12/12/2001                                  
*/

/**************************************************************************
$Id: cctrans_sample.p,v 1.4 2004/07/02 15:31:15 dan Exp $

  Author: Dan Yost <dan@tri8.com>
    Date: 12/12/2001

   USAGE: cctrans ( input-output temp-table_trExec,
                    input log_action-override, /* yes means do not actually
                                                  perform action specified
                                                  right now */
                  )
                  

  OUTPUT:  The trExec table is an input-output parameter and is used to
	   return the results.

  PREFACE:
    This program is provided as a sample of how one might use OneStep Charge
    to interface with a Progress database.  This program relies on a table
    called "cc-trans" which is assumed to exist in your database as the 
    table for permanent credit card transaction storage.  A second table called
    "cc-batch" is used to support automatic re-tries, though such processing
    would be done elsewhere as cc-batch records are merely set up here.  Also,
    consider defining functions to append the available error-status messages
    to the returned detail variables, as these are helpful in troubleshooting.

  DESCRIPTION:
    This program creates a cc-trans record from the supplied trExec record
    and executes the transaction as indicated by trExec.action, unless 
    action-override is true.  If action-override is true, the program exits
    after creating and assigning the cc-trans record.  Output is returned
    by synchronizing the cc-trans table with the trExec table.  The 
    cc-trans.cardNo field is not copied back to trExec, since it will be 
    encoded.  This allows the caller to continue seeing the full cardNo
    unencoded, and since trExec is a temp-table this should not be a 
    security concern.

    NOTE that the cc-batch and cc-trans tables used here are just examples
    of how one might use OSC and are not part of the actual OSC 
    infrastructure.

    Any program that executes CC transactions must establish a cc-batch
    record for each transaction.  The cc-batch record specifies what action
    will be performed using the data from the cc-trans record (purchase, 
    credit, etc.) and also controls automatic re-tries via the do-retry 
    field.  If cctrans is called with action-override = true, then no 
    cc-batch record is created, since cctrans will not be taking action on 
    the card.  If action-override is false, then cctrans will generate an 
    appropriate cc-batch record (this will be a "batch" of one card) and 
    update it as necessary.
    
    The actual execution of the transaction indicated by the trExec record
    is done by passing this record to the OSC engine, which will perform the 
    transaction.  After the OSC engine has been dispatched and returns, the 
    results are recorded in the cc-trans record. In the event that the results 
    cannot be recorded (due to a lock failure, etc) the trExec record is saved 
    to disk via panicSave().  If an error in locking occurs *before* the OSC
    engine dispatch, the dispatch will be cancelled and the cc-trans record 
    will be left alone.  Generally, this means that cc-trans.authStatus will 
    be CCS_NODO.  No panicSave is done in this case, since the dispatch was 
    never executed and no results must be saved.
    

  NOTES:
    * virtually all validation/filtering must be done before passing the
      trExec record to this program.  cctrans does almost no validation
      or filtering.

**************************************************************************/

           /*********************************************
           * COPYRIGHT Tri-8, Inc. ALL RIGHTS RESERVED. * 
           *********************************************/

/*************************** DEFINITIONS *********************************/
{tri8osc/oscapi}

def input-output parameter table for trExec.
def input parameter in-action-override as log no-undo.

def var tmp-cardNo as char no-undo.
def var is-credit as log no-undo.
def var exp-month as int no-undo.
def var exp-year as int no-undo.
def var use-ccp-name as char no-undo.
def var use-ccp-vrsn as char no-undo.
def var ccp-runpgm as char no-undo.
def var i as int no-undo.
def var ustr as char no-undo.


/*************************** MAIN BLOCK **********************************/


is-credit = no.

find trExec exclusive-lock no-wait no-error.

if error-status:error 
then return.            


/* cut unnecessary whitespace, which can cause problems elsewhere */
assign trExec.cardNo  = num(trExec.cardNo).
       trExec.expDate = trim(trExec.expDate).
       trExec.CHName = trim(trExec.CHName).
       trExec.CHAddress = trim(trExec.CHAddress).
       trExec.CHZip = trim(trExec.CHZip).
       trExec.CVV2 = trim(trExec.CVV2).

/* enable/disable actions here */
case trExec.action:
    when CCA_AUTHONLY then .
    when CCA_PURCHASE then .
    when CCA_CREDIT then .
    when CCA_VOICEAUTH then .
    otherwise do:
        assign trExec.authDetails  = trExec.authDetails
                                   + "invalid credit card action (or not "
				   + "appropriate for transaction *creation*)"
	       trExec.transNo = -1.
        return.
    end.
end case. /* in-action */


assign use-ccp-name = "OSC"
       use-ccp-vrsn = "2.0"
       ccp-runpgm = "tri8osc/oscengin.p".


/* create a cc-trans record */
CRBLOCK:
do transaction on error undo CRBLOCK, retry CRBLOCK:

    if retry
    then do:
        assign trExec.transNo     = -1
               trExec.authDetails = "Cannot create/assign cc-trans" no-error.
        return.
    end.
           
    create cc-trans.
    assign cc-trans.card-no      = trExec.cardNo  /* if you store CC numbers,
						  // you should wrap this
						  // with your encryption 
						  // method */

           /* you can add your own fields to trExec (found in tri8osc/oscapi)
	   // in order to pass in fields like "ar-entity," "cust-no," etc. */
	   /****
           cc-trans.cust-no      = trExec.cust-no
	   ****/

           cc-trans.exp-date     = trExec.expDate
           cc-trans.amount       = trExec.amount
           cc-trans.ch-name      = trExec.CHName
           cc-trans.ch-addr      = trExec.CHAddress
           cc-trans.ch-zip       = trExec.CHZip
           cc-trans.CVV2         = trExec.CVV2
           cc-trans.ccp-name     = use-ccp-name
           cc-trans.ccp-vrsn     = use-ccp-vrsn
           cc-trans.auth-status  = CCS_NODO
	   cc-trans.appr-code    = trExec.apprCode
           cc-trans.req-date     = today
           cc-trans.req-time     = time
           cc-trans.auth-date    = ?
           cc-trans.auth-time    = 0
           cc-trans.settle-date  = ?
           cc-trans.settle-time  = 0
           cc-trans.comment      = trExec.comment
           cc-trans.source       = trExec.source
           cc-trans.entered-by   = trExec.enteredBy
           cc-trans.trans-no     = 1234.  /* use a sequence 
					  // number (next-value) */

    /* multiple assignments for use of user-defined functions with key fields */
    assign cc-trans.card-no-mask = CCMask(trExec.cardNo).

end. /* CRBLOCK */

/* return if we're not taking action now */
if in-action-override = true 
then return.

/* at this point, since we *will* be taking action on the card, we need a
// cc-batch record */
CCBBLK:
do transaction:
    create cc-batch no-error.
    assign cc-batch.batch-no     = 999   /* use a sequence value */
           cc-batch.trans-no     = cc-trans.trans-No
           cc-batch.do-retry     = false
           cc-batch.num-tries    = 1
           cc-batch.submit-date  = today
           cc-batch.submit-time  = time
           cc-batch.action       = trExec.action
           no-error.
    
    if error-status:error
    then do:
        find current cc-trans exclusive-lock no-wait no-error.
        assign cc-trans.auth-details = cc-trans.auth-details
                                     + "ERROR creating/assigning cc-batch "
                                     + "record in ns/cctrans"
               cc-trans.auth-status  = CCS_SYSERR                       
               no-error.

	/* copy back cc-trans data to cc-exec here */

        return.       
    end. /* if error-status:error  */
end. /* CCBBLK */

/* dispatch the OSC engine, which handles the actual processing,
// according to the standard OSC API */
   
/* NOTE: in some cases a run statement can cause a "stop" condition without
// causing an error condition, so we use a little more cumbersome structure
// in order to catch such scenarios */
   
DSPBLOCK:
do on stop undo DSPBLOCK, retry DSPBLOCK
   on endkey undo DSPBLOCK, retry DSPBLOCK
   on error undo DSPBLOCK, retry DSPBLOCK transaction:
   
    /* handle stop/endkey/error */
    if retry 
    then do:
        find current trExec exclusive-lock no-wait no-error.
        assign trExec.authDetails = trExec.authDetails
                                  + "ERROR dispatching CCP controller: ~"" 
                                  + ccp-runpgm
               trExec.authStatus  = if trExec.authStatus <> CCS_SUBERR
				    then CCS_SYSERR
				    else CCS_SUBERR
               no-error.

	/* copy trExec fields back to cc-trans here, for permanent storage */

        find current cc-batch exclusive-lock no-wait no-error.

        assign cc-batch.do-retry = if trExec.authStatus <> CCS_SUBERR
				   then true
				   else false
	       no-error.

        if cc-batch.do-retry = true
	then assign cc-trans.card-no-pt = trExec.cardNo no-error.

	if trExec.authStatus = CCS_SUBERR
	then true. /* perhaps issue major warning/log */

        return.
    end. /* if retry */

    /* dispatch the CCP controller */
    run value(ccp-runpgm) ( input-output table trExec) no-error.
    
    if error-status:error then undo DSPBLOCK, retry DSPBLOCK.
    
    find trExec exclusive-lock no-error.
    if error-status:error
    then do:
	/* msg/log here */
        return.
    end. /* if error-status:error */

    /* lock the current cc-trans exclusively with no blocking */
    find current cc-trans exclusive-lock no-wait no-error.
    if error-status:error
    then do:
        assign trExec.authStatus = CCS_SUBERR
               trExec.authDetails = trExec.authDetails 
                                  + "ERROR: can't lock cc-trans, aborting with "
                                  + "panicSave"
               no-error.

        panicSave(input table trExec, -1).
        return.
    end. /* if error-status:error */

    /* record the results of the transaction 
    // (except for the unencoded cardNo field) */
    buffer-copy trExec except trExec.cardNo to cc-trans no-error.
    
    if error-status:error
    then do:
	assign trExec.authDetails = trExec.authDetails
				  + "ERROR: can't buffer-copy trExec, panicSave"
	       trExec.authStatus  = CCS_SUBERR
	       no-error.

        panicSave(input table trExec, -1).
    end. /* if error-status:error in copying trExec */

    /* lock the current cc-batch exclusively with no blocking */
    find current cc-batch exclusive-lock no-wait no-error.
    if error-status:error
    then do:
        assign trExec.authDetails = trExec.authDetails 
                                  + "ERROR: can't lock cc-batch, aborting "
                                  + "with panicSave"
               trExec.authStatus  = CCS_SUBERR
               no-error.

        buffer-copy trExec except trExec.cardNo to cc-trans no-error.

        panicSave(input table trExec, -1).
        return.
    end. /* if error-status:error */

    assign cc-batch.num-tries = cc-batch.num-tries + 1.

    /* update cc-batch and cc-trans if necessary */
    if trExec.authStatus = CCS_SYSERR
    then do:
	assign cc-batch.do-retry = true
	       cc-trans.card-no-pt = trExec.cardNo 
               no-error.
    end. /* if trExec.authStatus = CCS_SYSERR */
end.  /* DSPBLOCK */
